////////////////////////////////////////////////////////////////////////
//                                                                    //
// MicrOTP: A tool for computing one time passwords for S/Key logins. //
//                                                                    //
// (C) 2008 Tom Bramer                                                //
//                                                                    //
// This program is free software; you can redistribute it and/or      //
// modify it under the terms of the GNU General Public License        //
// as published by the Free Software Foundation; either version 2     //
// of the License, or (at your option) any later version.             //
//                                                                    //
// This program is distributed in the hope that it will be useful,    //
// but WITHOUT ANY WARRANTY; without even the implied warranty of     //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      //
// GNU General Public License for more details.                       //
//                                                                    //
// You should have received a copy of the GNU General Public License  //
// along with this program; if not, write to the Free Software        //
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA      //
// 02110-1301, USA.                                                   //
//                                                                    //
////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <fv/config.h>
#endif

#include <string>
#include <vector>
#include <windows.h>
#include <tchar.h>
#include <fv/defs.hpp>
#include <fv/types.hpp>
#include <fv/hashiface.hpp>
#include <fv/build.hpp>
#include <fv/version.hpp>
#include <stdio.h>
#include <cstring>
#include <cstdlib>
#include "common.hpp"

USING_NAMESPACE(PU::FV);

typedef unsigned short Word;
typedef unsigned long DWord;

#ifndef BIG_ENDIAN  /* Assume LITTLE_ENDIAN */
#define ByteReverse(buffer, len)
#else

void ByteReverse(Byte* buffer, DWord len)
{
    do {
        DWord temp =
        (((DWord) buffer[0] << 8) | buffer[1]) << 16 |
        (((DWord) buffer[2] << 8) | buffer[3]);
        *(DWord*) buffer = temp;
        buffer += 4;
    } while (--len);
    return;
}

#endif

class CMDBase : public CHashAlgorithm {
public:
    virtual const char* GetAlgName() = 0;
    virtual const char* GetImpAuthor() = 0;
    virtual int  GetImpVersion() = 0;
    virtual const char* GetAdditionalInfo() = 0;
    virtual void Transform() = 0;
    static inline DWord ROTL(DWord X, DWord Y) { return(X << Y) | (X >> (32 - Y));}

    virtual void Initialize() 
    {
        m_nResult[0] = 0x67452301;
        m_nResult[1] = 0xefcdab89;
        m_nResult[2] = 0x98badcfe;
        m_nResult[3] = 0x10325476;

        m_nBits[0] = 0;
        m_nBits[1] = 0;
        return;
    }

    virtual void DoStep(Byte* pBuffer, BufferLength nLength)
    {
        DWord temp;

        temp = m_nBits[0];
        if ((m_nBits[0] = temp + ((DWord) nLength << 3)) < temp)
            m_nBits[1]++;
        m_nBits[1] += nLength >> 29;
        temp = (temp >> 3) & 63;

        if (temp) {  /* Handle buffered data */
            Byte* bp = m_aBuffer + temp;
            temp = 64 - temp;

            if (nLength < temp) {
                memcpy(bp, pBuffer, nLength);
                return;
            }

            memcpy(bp, pBuffer, temp);
            ByteReverse(m_aBuffer, 16);
            Transform();
            pBuffer += temp;
            nLength -= temp;
        }

        while (nLength >= 64) {
            memcpy(m_aBuffer, pBuffer, 64);
            ByteReverse(m_aBuffer, 16);
            Transform();
            pBuffer += 64;
            nLength -= 64;
        }

        memcpy(m_aBuffer, pBuffer, nLength);
    }

    virtual void Finalize()
    {
        DWord count;
        Byte* bp;

        count = (m_nBits[0] >> 3) & 0x3F;

        bp = m_aBuffer + count;
        *bp++ = 0x80;

        count = 63 - count;

        if (count < 8) {
            memset(bp, 0, count);
            ByteReverse(m_aBuffer, 16);
            Transform();
            memset(m_aBuffer, 0, 56);
        } else {
            memset(bp, 0, count - 8);
        }

        ByteReverse(m_aBuffer, 14);

        ((DWord*)m_aBuffer)[14] = m_nBits[0];
        ((DWord*)m_aBuffer)[15] = m_nBits[1];

        Transform();
        ByteReverse((Byte*)m_nResult, 4);
        memset(m_aBuffer, 0, 64);
        return;
    }

	virtual int GetStringResult(char* pBuffer, int nBufferSize)
    {
        if(nBufferSize == 33) {
			HashToHexStr(pBuffer, (Byte*)m_nResult, 16);
		}

		return 33;
    }

    virtual int GetByteResult(Byte* pBuffer, int nBufferSize)
    {
        if (nBufferSize == 16) {
            memcpy(pBuffer, (Byte*)m_nResult, 16);
        }
        return 16;
    }

protected:
    DWord m_nResult[4];
    DWord m_nBits[2];
    Byte  m_aBuffer[64];
};


class CMD4Algorithm : public CMDBase {
public:
    virtual const char* GetAlgName() { return "MD4";}
    virtual const char* GetImpAuthor() { return "Programming Unlimited";}
    virtual int  GetImpVersion() { return 1000000;}
    virtual const char* GetAdditionalInfo() { return "Ron Rivest's MD4 algorithm.  Produces 128-bit output like MD5, but is "
                                                     "less secure.  Should be used only for backward-compatability purposes.";}

    static inline DWord F(DWord X, DWord Y, DWord Z) { return(X & Y) | (~X & Z);}
    static inline DWord G(DWord X, DWord Y, DWord Z) { return(X & Y) | (X & Z) | (Y & Z);}
    static inline DWord H(DWord X, DWord Y, DWord Z) { return X ^ Y ^ Z;}

    static const DWord S[12];

    static inline void FF(DWord& A, DWord B, DWord C, DWord D, DWord X, DWord S)
    {
        A += F(B, C, D) + X;
        A = ROTL(A, S);
    }

    static inline void GG(DWord& A, DWord B, DWord C, DWord D, DWord X, DWord S)
    {
        A += G(B, C, D) + X + 0x5A827999;
        A = ROTL(A, S);
    }

    static inline void HH(DWord& A, DWord B, DWord C, DWord D, DWord X, DWord S)
    {
        A += H(B, C, D) + X + 0x6ED9EBA1;
        A = ROTL(A, S);
    }

    virtual void Transform()
    {
        register DWord a, b, c, d;
        register DWord* pBuffer = reinterpret_cast<DWord*>(m_aBuffer);

        a = m_nResult[0];
        b = m_nResult[1];
        c = m_nResult[2];
        d = m_nResult[3];

        FF(a, b, c, d, pBuffer[0 ], S[0 ]); /*  1 */
        FF(d, a, b, c, pBuffer[1 ], S[1 ]); /*  2 */
        FF(c, d, a, b, pBuffer[2 ], S[2 ]); /*  3 */
        FF(b, c, d, a, pBuffer[3 ], S[3 ]); /*  4 */
        FF(a, b, c, d, pBuffer[4 ], S[0 ]); /*  5 */
        FF(d, a, b, c, pBuffer[5 ], S[1 ]); /*  6 */
        FF(c, d, a, b, pBuffer[6 ], S[2 ]); /*  7 */
        FF(b, c, d, a, pBuffer[7 ], S[3 ]); /*  8 */
        FF(a, b, c, d, pBuffer[8 ], S[0 ]); /*  9 */
        FF(d, a, b, c, pBuffer[9 ], S[1 ]); /* 10 */
        FF(c, d, a, b, pBuffer[10], S[2 ]); /* 11 */
        FF(b, c, d, a, pBuffer[11], S[3 ]); /* 12 */
        FF(a, b, c, d, pBuffer[12], S[0 ]); /* 13 */
        FF(d, a, b, c, pBuffer[13], S[1 ]); /* 14 */
        FF(c, d, a, b, pBuffer[14], S[2 ]); /* 15 */
        FF(b, c, d, a, pBuffer[15], S[3 ]); /* 16 */

        GG(a, b, c, d, pBuffer[0 ], S[4 ]); /* 17 */
        GG(d, a, b, c, pBuffer[4 ], S[5 ]); /* 18 */
        GG(c, d, a, b, pBuffer[8 ], S[6 ]); /* 19 */
        GG(b, c, d, a, pBuffer[12], S[7 ]); /* 20 */
        GG(a, b, c, d, pBuffer[1 ], S[4 ]); /* 21 */
        GG(d, a, b, c, pBuffer[5 ], S[5 ]); /* 22 */
        GG(c, d, a, b, pBuffer[9 ], S[6 ]); /* 23 */
        GG(b, c, d, a, pBuffer[13], S[7 ]); /* 24 */
        GG(a, b, c, d, pBuffer[2 ], S[4 ]); /* 25 */
        GG(d, a, b, c, pBuffer[6 ], S[5 ]); /* 26 */
        GG(c, d, a, b, pBuffer[10], S[6 ]); /* 27 */
        GG(b, c, d, a, pBuffer[14], S[7 ]); /* 28 */
        GG(a, b, c, d, pBuffer[3 ], S[4 ]); /* 29 */
        GG(d, a, b, c, pBuffer[7 ], S[5 ]); /* 30 */
        GG(c, d, a, b, pBuffer[11], S[6 ]); /* 31 */
        GG(b, c, d, a, pBuffer[15], S[7 ]); /* 32 */

        HH(a, b, c, d, pBuffer[0 ], S[8 ]); /* 33 */
        HH(d, a, b, c, pBuffer[8 ], S[9 ]); /* 34 */
        HH(c, d, a, b, pBuffer[4 ], S[10]); /* 35 */
        HH(b, c, d, a, pBuffer[12], S[11]); /* 36 */
        HH(a, b, c, d, pBuffer[2 ], S[8 ]); /* 37 */
        HH(d, a, b, c, pBuffer[10], S[9 ]); /* 38 */
        HH(c, d, a, b, pBuffer[6 ], S[10]); /* 39 */
        HH(b, c, d, a, pBuffer[14], S[11]); /* 40 */
        HH(a, b, c, d, pBuffer[1 ], S[8 ]); /* 41 */
        HH(d, a, b, c, pBuffer[9 ], S[9 ]); /* 42 */
        HH(c, d, a, b, pBuffer[5 ], S[10]); /* 43 */
        HH(b, c, d, a, pBuffer[13], S[11]); /* 44 */
        HH(a, b, c, d, pBuffer[3 ], S[8 ]); /* 45 */
        HH(d, a, b, c, pBuffer[11], S[9 ]); /* 46 */
        HH(c, d, a, b, pBuffer[7 ], S[10]); /* 47 */
        HH(b, c, d, a, pBuffer[15], S[11]); /* 48 */

        m_nResult[0] += a;
        m_nResult[1] += b;
        m_nResult[2] += c;
        m_nResult[3] += d;

        return;
    }

};

const DWord CMD4Algorithm::S[12] =
{
    3, 7, 11, 19, 3, 5, 9, 13, 3, 9, 11, 15
};


class CMD5Algorithm : public CMDBase {
public:
    virtual const char* GetAlgName() { return "MD5";}
    virtual const char* GetImpAuthor() { return "Programming Unlimited";}
    virtual int  GetImpVersion() { return 1000000;}
    virtual const char* GetAdditionalInfo() { return "Ron Rivest's MD5 algorithm.  Produces 128-bit output.  This is "
                                                     "the successor to MD4, a similar, yet less secure (less collsiion-resistant) "
                                                     "algorithm.";}


    static inline DWord F(DWord X, DWord Y, DWord Z) { return Z ^ (X & (Y ^ Z));}
    static inline DWord G(DWord X, DWord Y, DWord Z) { return F(Z, X, Y);}
    static inline DWord H(DWord X, DWord Y, DWord Z) { return X ^ Y ^ Z;}
    static inline DWord I(DWord X, DWord Y, DWord Z) { return Y ^ (X | ~Z);}

    static inline void FF(DWord& A, DWord B, DWord C, DWord D, DWord X, DWord S, DWord AC)
    {
        A += F(B, C, D) + X + AC;
        A = ROTL(A, S);
        A += B;
    }

    static inline void GG(DWord& A, DWord B, DWord C, DWord D, DWord X, DWord S, DWord AC)
    {
        A += G(B, C, D) + X + AC;
        A = ROTL(A, S);
        A += B;
    }

    static inline void HH(DWord& A, DWord B, DWord C, DWord D, DWord X, DWord S, DWord AC)
    {
        A += H(B, C, D) + X + AC;
        A = ROTL(A, S);
        A += B;
    }

    static inline void II(DWord& A, DWord B, DWord C, DWord D, DWord X, DWord S, DWord AC)
    {
        A += I(B, C, D) + X + AC;
        A = ROTL(A, S);
        A += B;
    }

    virtual void Transform()
    {
        register DWord a, b, c, d;
        register DWord* pBuffer = reinterpret_cast<DWord*>(m_aBuffer);

        a = m_nResult[0];
        b = m_nResult[1];
        c = m_nResult[2];
        d = m_nResult[3];

        FF(a, b, c, d, pBuffer[0 ],  7,  0xD76AA478); /*  1 */
        FF(d, a, b, c, pBuffer[1 ], 12,  0xE8C7B756); /*  2 */
        FF(c, d, a, b, pBuffer[2 ], 17,  0x242070DB); /*  3 */
        FF(b, c, d, a, pBuffer[3 ], 22,  0xC1BDCEEE); /*  4 */
        FF(a, b, c, d, pBuffer[4 ],  7,  0xF57C0FAF); /*  5 */
        FF(d, a, b, c, pBuffer[5 ], 12,  0x4787C62A); /*  6 */
        FF(c, d, a, b, pBuffer[6 ], 17,  0xA8304613); /*  7 */
        FF(b, c, d, a, pBuffer[7 ], 22,  0xFD469501); /*  8 */
        FF(a, b, c, d, pBuffer[8 ],  7,  0x698098D8); /*  9 */
        FF(d, a, b, c, pBuffer[9 ], 12,  0x8B44F7AF); /* 10 */
        FF(c, d, a, b, pBuffer[10], 17,  0xFFFF5BB1); /* 11 */
        FF(b, c, d, a, pBuffer[11], 22,  0x895CD7BE); /* 12 */
        FF(a, b, c, d, pBuffer[12],  7,  0x6B901122); /* 13 */
        FF(d, a, b, c, pBuffer[13], 12,  0xFD987193); /* 14 */
        FF(c, d, a, b, pBuffer[14], 17,  0xA679438E); /* 15 */
        FF(b, c, d, a, pBuffer[15], 22,  0x49B40821); /* 16 */

        GG(a, b, c, d, pBuffer[1 ],  5,  0xF61E2562); /* 17 */
        GG(d, a, b, c, pBuffer[6 ],  9,  0xC040B340); /* 18 */
        GG(c, d, a, b, pBuffer[11], 14,  0x265E5A51); /* 19 */
        GG(b, c, d, a, pBuffer[0 ], 20,  0xE9B6C7AA); /* 20 */
        GG(a, b, c, d, pBuffer[5 ],  5,  0xD62F105D); /* 21 */
        GG(d, a, b, c, pBuffer[10],  9,  0x02441453); /* 22 */
        GG(c, d, a, b, pBuffer[15], 14,  0xD8A1E681); /* 23 */
        GG(b, c, d, a, pBuffer[4 ], 20,  0xE7D3FBC8); /* 24 */
        GG(a, b, c, d, pBuffer[9 ],  5,  0x21E1CDE6); /* 25 */
        GG(d, a, b, c, pBuffer[14],  9,  0xC33707D6); /* 26 */
        GG(c, d, a, b, pBuffer[3 ], 14,  0xF4D50D87); /* 27 */
        GG(b, c, d, a, pBuffer[8 ], 20,  0x455A14ED); /* 28 */
        GG(a, b, c, d, pBuffer[13],  5,  0xA9E3E905); /* 29 */
        GG(d, a, b, c, pBuffer[2 ],  9,  0xFCEFA3F8); /* 30 */
        GG(c, d, a, b, pBuffer[7 ], 14,  0x676F02D9); /* 31 */
        GG(b, c, d, a, pBuffer[12], 20,  0x8D2A4C8A); /* 32 */

        HH(a, b, c, d, pBuffer[5 ],  4,  0xFFFA3942); /* 33 */
        HH(d, a, b, c, pBuffer[8 ], 11,  0x8771F681); /* 34 */
        HH(c, d, a, b, pBuffer[11], 16,  0x6D9D6122); /* 35 */
        HH(b, c, d, a, pBuffer[14], 23,  0xFDE5380C); /* 36 */
        HH(a, b, c, d, pBuffer[1 ],  4,  0xA4BEEA44); /* 37 */
        HH(d, a, b, c, pBuffer[4 ], 11,  0x4BDECFA9); /* 38 */
        HH(c, d, a, b, pBuffer[7 ], 16,  0xF6BB4B60); /* 39 */
        HH(b, c, d, a, pBuffer[10], 23,  0xBEBFBC70); /* 40 */
        HH(a, b, c, d, pBuffer[13],  4,  0x289B7EC6); /* 41 */
        HH(d, a, b, c, pBuffer[0 ], 11,  0xEAA127FA); /* 42 */
        HH(c, d, a, b, pBuffer[3 ], 16,  0xD4EF3085); /* 43 */
        HH(b, c, d, a, pBuffer[6 ], 23,  0x04881D05); /* 44 */
        HH(a, b, c, d, pBuffer[9 ],  4,  0xD9D4D039); /* 45 */
        HH(d, a, b, c, pBuffer[12], 11,  0xE6DB99E5); /* 46 */
        HH(c, d, a, b, pBuffer[15], 16,  0x1FA27CF8); /* 47 */
        HH(b, c, d, a, pBuffer[2 ], 23,  0xC4AC5665); /* 48 */

        II(a, b, c, d, pBuffer[0 ],  6,  0xF4292244); /* 49 */
        II(d, a, b, c, pBuffer[7 ], 10,  0x432AFF97); /* 50 */
        II(c, d, a, b, pBuffer[14], 15,  0xAB9423A7); /* 51 */
        II(b, c, d, a, pBuffer[5 ], 21,  0xFC93A039); /* 52 */
        II(a, b, c, d, pBuffer[12],  6,  0x655B59C3); /* 53 */
        II(d, a, b, c, pBuffer[3 ], 10,  0x8F0CCC92); /* 54 */
        II(c, d, a, b, pBuffer[10], 15,  0xFFEFF47D); /* 55 */
        II(b, c, d, a, pBuffer[1 ], 21,  0x85845DD1); /* 56 */
        II(a, b, c, d, pBuffer[8 ],  6,  0x6FA87E4F); /* 57 */
        II(d, a, b, c, pBuffer[15], 10,  0xFE2CE6E0); /* 58 */
        II(c, d, a, b, pBuffer[6 ], 15,  0xA3014314); /* 59 */
        II(b, c, d, a, pBuffer[13], 21,  0x4E0811A1); /* 60 */
        II(a, b, c, d, pBuffer[4 ],  6,  0xF7537E82); /* 61 */
        II(d, a, b, c, pBuffer[11], 10,  0xBD3AF235); /* 62 */
        II(c, d, a, b, pBuffer[2 ], 15,  0x2AD7D2BB); /* 63 */
        II(b, c, d, a, pBuffer[9 ], 21,  0xEB86D391); /* 64 */

        m_nResult[0] += a;
        m_nResult[1] += b;
        m_nResult[2] += c;
        m_nResult[3] += d;

        return;
    }

};
