////////////////////////////////////////////////////////////////////////
//                                                                    //
// MicrOTP: A tool for computing one time passwords for S/Key logins. //
//                                                                    //
// (C) 2008 Tom Bramer                                                //
//                                                                    //
// This program is free software; you can redistribute it and/or      //
// modify it under the terms of the GNU General Public License        //
// as published by the Free Software Foundation; either version 2     //
// of the License, or (at your option) any later version.             //
//                                                                    //
// This program is distributed in the hope that it will be useful,    //
// but WITHOUT ANY WARRANTY; without even the implied warranty of     //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      //
// GNU General Public License for more details.                       //
//                                                                    //
// You should have received a copy of the GNU General Public License  //
// along with this program; if not, write to the Free Software        //
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA      //
// 02110-1301, USA.                                                   //
//                                                                    //
////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <fv/config.h>
#endif

#include <string>
#include <vector>
#include <windows.h>
#include <tchar.h>
#include <stdio.h>
#include <fv/defs.hpp>
#include <fv/types.hpp>
#include <fv/hashiface.hpp>
#include <fv/build.hpp>
#include <fv/version.hpp>
#include "common.hpp"

USING_NAMESPACE(PU::FV);

typedef unsigned short Word;
typedef unsigned long DWord;



#ifndef BIG_ENDIAN  /* Assume BIG_ENDIAN */

template <typename T>
class ByteReverse {
public:
    static inline void Do(Byte* buffer, DWord len) { return ;}
};

#else

template <typename T>
class ByteReverse {
};

template <>
class ByteReverse<DWord> {
public:
    static inline void Do(Byte* buffer, DWord len)
    {
        do {
            DWord temp =
            (((DWord) buffer[0] << 8) | buffer[1]) << 16 |
            (((DWord) buffer[2] << 8) | buffer[3]);
            *(DWord*) buffer = temp;
            buffer += 4;
        } while (--len);
        return;
    }
};

#endif



class CRMDFuncBase {
public:
    static inline DWord ROTL(DWord X, DWord Y) 
    { return(X << Y) | (X >> (32 - Y));}
    static inline DWord ROTR(DWord X, DWord Y) 
    { return(X >> Y) | (X << (32 - Y));}
    static inline DWord F(DWord X, DWord Y, DWord Z) { return X ^ Y ^ Z;}
    static inline DWord G(DWord X, DWord Y, DWord Z) { return(X & Y) | (~X & Z);}
    static inline DWord H(DWord X, DWord Y, DWord Z) { return(X | ~Y) ^ Z;}
    static inline DWord I(DWord X, DWord Y, DWord Z) { return(X & Z) | (Y & ~Z);}
    static inline DWord J(DWord X, DWord Y, DWord Z) { return X ^ (Y | ~Z);}
};



template <int HashSize>
class CRMDBase : public CHashAlgorithm {
public:
	virtual const char* GetAlgName() = 0;
	virtual const char* GetImpAuthor() = 0;
	virtual int  GetImpVersion() = 0;
	virtual const char* GetAdditionalInfo() = 0;
	virtual void Transform() = 0;
    virtual void Initialize() = 0;

    CRMDBase() { m_pBuffer = reinterpret_cast<DWord*>(m_aBuffer);}

    virtual void DoStep(Byte* pBuffer, BufferLength nLength)
    {
        DWord temp;

        temp = m_nBits[0];
        if ((m_nBits[0] = temp + ((DWord) nLength << 3)) < temp)
            m_nBits[1]++;

        m_nBits[1] += nLength >> 29;
        temp = (temp >> 3) & 0x3F;

        if (temp) {  /* Handle buffered data */
            Byte* bp = m_aBuffer + temp;
            temp = 63 - temp;

            if (nLength < temp) {
                memcpy(bp, pBuffer, nLength);
                return;
            }

            memcpy(bp, pBuffer, temp);
            ByteReverse<DWord>::Do(m_aBuffer, 16);
            Transform();
            pBuffer += temp;
            nLength -= temp;
        }

        while (nLength >= 64) {
            memcpy(m_aBuffer, pBuffer, 64);
            ByteReverse<DWord>::Do(m_aBuffer, 16);
            Transform();
            pBuffer += 64;
            nLength -= 64;
        }

        memcpy(m_aBuffer, pBuffer, nLength);
    }

    virtual void Finalize()
    {
        DWord count;
        Byte* bp;

        count = (m_nBits[0] >> 3) & 63;

        bp = m_aBuffer + count;
        *bp++ = 0x80;

        count = 63 - count;

        if (count < 8) {
            memset(bp, 0, count);
            ByteReverse<DWord>::Do(m_aBuffer, 16);
            Transform();
            memset(m_aBuffer, 0, 56);
        } else {
            memset(bp, 0, count - 8);
        }

        ((DWord*)m_aBuffer)[14] = m_nBits[0];
        ((DWord*)m_aBuffer)[15] = m_nBits[1];
        ByteReverse<DWord>::Do(m_aBuffer, 14);

        Transform();
        ByteReverse<DWord>::Do((Byte*)m_nResult, HashSize);
        memset(m_aBuffer, 0, 64);
        return;
    }

	virtual int GetStringResult(char* pBuffer, int nBufferSize)
	{
		if(nBufferSize == (8 * HashSize) + 1) {
			HashToHexStr(pBuffer, (Byte*)m_nResult, 4 * HashSize);
		}

		return (8 * HashSize) + 1;
	}

    virtual int GetByteResult(Byte* pBuffer, int nBufferSize)
    {
        if (nBufferSize == 4 * HashSize) {
            memcpy(pBuffer, (Byte*)m_nResult, 4 * HashSize);
        }
        return 4 * HashSize;
    }

protected:
    DWord  m_nResult[HashSize];
    DWord  m_nBits[2];
    Byte   m_aBuffer[64];
    DWord* m_pBuffer;
};







class CRMD160Base : public CRMDFuncBase {
public:
    static inline void FF(DWord& A, DWord B, DWord& C, DWord D, DWord E, DWord X, DWord S)
    {
        A += F(B, C, D) + X;
        A = ROTL(A, S) + E;
        C = ROTL(C, 10);
    }

    static inline void GG(DWord& A, DWord B, DWord& C, DWord D, DWord E, DWord X, DWord S)
    {
        A += G(B, C, D) + X + 0x5a827999;
        A = ROTL(A, S) + E;
        C = ROTL(C, 10);
    }

    static inline void HH(DWord& A, DWord B, DWord& C, DWord D, DWord E, DWord X, DWord S)
    {
        A += H(B, C, D) + X + 0x6ed9eba1;
        A = ROTL(A, S) + E;
        C = ROTL(C, 10);
    }

    static inline void II(DWord& A, DWord B, DWord& C, DWord D, DWord E, DWord X, DWord S)
    {
        A += I(B, C, D) + X + 0x8f1bbcdc;
        A = ROTL(A, S) + E;
        C = ROTL(C, 10);
    }

    static inline void JJ(DWord& A, DWord B, DWord& C, DWord D, DWord E, DWord X, DWord S)
    {
        A += J(B, C, D) + X + 0xa953fd4e;
        A = ROTL(A, S) + E;
        C = ROTL(C, 10);
    }

    static inline void FFF(DWord& A, DWord B, DWord& C, DWord D, DWord E, DWord X, DWord S)
    {
        A += F(B, C, D) + X;
        A = ROTL(A, S) + E;
        C = ROTL(C, 10);
    }

    static inline void GGG(DWord& A, DWord B, DWord& C, DWord D, DWord E, DWord X, DWord S)
    {
        A += G(B, C, D) + X + 0x7a6d76e9;
        A = ROTL(A, S) + E;
        C = ROTL(C, 10);
    }

    static inline void HHH(DWord& A, DWord B, DWord& C, DWord D, DWord E, DWord X, DWord S)
    {
        A += H(B, C, D) + X + 0x6d703ef3;
        A = ROTL(A, S) + E;
        C = ROTL(C, 10);
    }

    static inline void III(DWord& A, DWord B, DWord& C, DWord D, DWord E, DWord X, DWord S)
    {
        A += I(B, C, D) + X + 0x5c4dd124;
        A = ROTL(A, S) + E;
        C = ROTL(C, 10);
    }

    static inline void JJJ(DWord& A, DWord B, DWord& C, DWord D, DWord E, DWord X, DWord S)
    {
        A += J(B, C, D) + X + 0x50a28be6;
        A = ROTL(A, S) + E;
        C = ROTL(C, 10);
    }

    static inline void FFR(DWord& a, DWord& b, DWord& c, DWord& d, DWord& e, DWord* pBuffer)
    {
        FF(a, b, c, d, e, pBuffer[ 0], 11);
        FF(e, a, b, c, d, pBuffer[ 1], 14);
        FF(d, e, a, b, c, pBuffer[ 2], 15);
        FF(c, d, e, a, b, pBuffer[ 3], 12);
        FF(b, c, d, e, a, pBuffer[ 4],  5);
        FF(a, b, c, d, e, pBuffer[ 5],  8);
        FF(e, a, b, c, d, pBuffer[ 6],  7);
        FF(d, e, a, b, c, pBuffer[ 7],  9);
        FF(c, d, e, a, b, pBuffer[ 8], 11);
        FF(b, c, d, e, a, pBuffer[ 9], 13);
        FF(a, b, c, d, e, pBuffer[10], 14);
        FF(e, a, b, c, d, pBuffer[11], 15);
        FF(d, e, a, b, c, pBuffer[12],  6);
        FF(c, d, e, a, b, pBuffer[13],  7);
        FF(b, c, d, e, a, pBuffer[14],  9);
        FF(a, b, c, d, e, pBuffer[15],  8);
    }

    static inline void GGR(DWord& a, DWord& b, DWord& c, DWord& d, DWord& e, DWord* pBuffer)
    {
        GG(e, a, b, c, d, pBuffer[ 7],  7);
        GG(d, e, a, b, c, pBuffer[ 4],  6);
        GG(c, d, e, a, b, pBuffer[13],  8);
        GG(b, c, d, e, a, pBuffer[ 1], 13);
        GG(a, b, c, d, e, pBuffer[10], 11);
        GG(e, a, b, c, d, pBuffer[ 6],  9);
        GG(d, e, a, b, c, pBuffer[15],  7);
        GG(c, d, e, a, b, pBuffer[ 3], 15);
        GG(b, c, d, e, a, pBuffer[12],  7);
        GG(a, b, c, d, e, pBuffer[ 0], 12);
        GG(e, a, b, c, d, pBuffer[ 9], 15);
        GG(d, e, a, b, c, pBuffer[ 5],  9);
        GG(c, d, e, a, b, pBuffer[ 2], 11);
        GG(b, c, d, e, a, pBuffer[14],  7);
        GG(a, b, c, d, e, pBuffer[11], 13);
        GG(e, a, b, c, d, pBuffer[ 8], 12);
    }

    static inline void HHR(DWord& a, DWord& b, DWord& c, DWord& d, DWord& e, DWord* pBuffer)
    {
        HH(d, e, a, b, c, pBuffer[ 3], 11);
        HH(c, d, e, a, b, pBuffer[10], 13);
        HH(b, c, d, e, a, pBuffer[14],  6);
        HH(a, b, c, d, e, pBuffer[ 4],  7);
        HH(e, a, b, c, d, pBuffer[ 9], 14);
        HH(d, e, a, b, c, pBuffer[15],  9);
        HH(c, d, e, a, b, pBuffer[ 8], 13);
        HH(b, c, d, e, a, pBuffer[ 1], 15);
        HH(a, b, c, d, e, pBuffer[ 2], 14);
        HH(e, a, b, c, d, pBuffer[ 7],  8);
        HH(d, e, a, b, c, pBuffer[ 0], 13);
        HH(c, d, e, a, b, pBuffer[ 6],  6);
        HH(b, c, d, e, a, pBuffer[13],  5);
        HH(a, b, c, d, e, pBuffer[11], 12);
        HH(e, a, b, c, d, pBuffer[ 5],  7);
        HH(d, e, a, b, c, pBuffer[12],  5);
    }

    static inline void IIR(DWord& a, DWord& b, DWord& c, DWord& d, DWord& e, DWord* pBuffer)
    {
        II(c, d, e, a, b, pBuffer[ 1], 11);
        II(b, c, d, e, a, pBuffer[ 9], 12);
        II(a, b, c, d, e, pBuffer[11], 14);
        II(e, a, b, c, d, pBuffer[10], 15);
        II(d, e, a, b, c, pBuffer[ 0], 14);
        II(c, d, e, a, b, pBuffer[ 8], 15);
        II(b, c, d, e, a, pBuffer[12],  9);
        II(a, b, c, d, e, pBuffer[ 4],  8);
        II(e, a, b, c, d, pBuffer[13],  9);
        II(d, e, a, b, c, pBuffer[ 3], 14);
        II(c, d, e, a, b, pBuffer[ 7],  5);
        II(b, c, d, e, a, pBuffer[15],  6);
        II(a, b, c, d, e, pBuffer[14],  8);
        II(e, a, b, c, d, pBuffer[ 5],  6);
        II(d, e, a, b, c, pBuffer[ 6],  5);
        II(c, d, e, a, b, pBuffer[ 2], 12);
    }

    static inline void JJR(DWord& a, DWord& b, DWord& c, DWord& d, DWord& e, DWord* pBuffer)
    {
        JJ(b, c, d, e, a, pBuffer[ 4],  9);
        JJ(a, b, c, d, e, pBuffer[ 0], 15);
        JJ(e, a, b, c, d, pBuffer[ 5],  5);
        JJ(d, e, a, b, c, pBuffer[ 9], 11);
        JJ(c, d, e, a, b, pBuffer[ 7],  6);
        JJ(b, c, d, e, a, pBuffer[12],  8);
        JJ(a, b, c, d, e, pBuffer[ 2], 13);
        JJ(e, a, b, c, d, pBuffer[10], 12);
        JJ(d, e, a, b, c, pBuffer[14],  5);
        JJ(c, d, e, a, b, pBuffer[ 1], 12);
        JJ(b, c, d, e, a, pBuffer[ 3], 13);
        JJ(a, b, c, d, e, pBuffer[ 8], 14);
        JJ(e, a, b, c, d, pBuffer[11], 11);
        JJ(d, e, a, b, c, pBuffer[ 6],  8);
        JJ(c, d, e, a, b, pBuffer[15],  5);
        JJ(b, c, d, e, a, pBuffer[13],  6);
    }

    static inline void FFFR(DWord& a, DWord& b, DWord& c, DWord& d, DWord& e, DWord* pBuffer)
    {
        FFF(b, c, d, e, a, pBuffer[12],  8);
        FFF(a, b, c, d, e, pBuffer[15],  5);
        FFF(e, a, b, c, d, pBuffer[10], 12);
        FFF(d, e, a, b, c, pBuffer[ 4],  9);
        FFF(c, d, e, a, b, pBuffer[ 1], 12);
        FFF(b, c, d, e, a, pBuffer[ 5],  5);
        FFF(a, b, c, d, e, pBuffer[ 8], 14);
        FFF(e, a, b, c, d, pBuffer[ 7],  6);
        FFF(d, e, a, b, c, pBuffer[ 6],  8);
        FFF(c, d, e, a, b, pBuffer[ 2], 13);
        FFF(b, c, d, e, a, pBuffer[13],  6);
        FFF(a, b, c, d, e, pBuffer[14],  5);
        FFF(e, a, b, c, d, pBuffer[ 0], 15);
        FFF(d, e, a, b, c, pBuffer[ 3], 13);
        FFF(c, d, e, a, b, pBuffer[ 9], 11);
        FFF(b, c, d, e, a, pBuffer[11], 11);
    }

    static inline void GGGR(DWord& a, DWord& b, DWord& c, DWord& d, DWord& e, DWord* pBuffer)
    {
        GGG(c, d, e, a, b, pBuffer[ 8], 15);
        GGG(b, c, d, e, a, pBuffer[ 6],  5);
        GGG(a, b, c, d, e, pBuffer[ 4],  8);
        GGG(e, a, b, c, d, pBuffer[ 1], 11);
        GGG(d, e, a, b, c, pBuffer[ 3], 14);
        GGG(c, d, e, a, b, pBuffer[11], 14);
        GGG(b, c, d, e, a, pBuffer[15],  6);
        GGG(a, b, c, d, e, pBuffer[ 0], 14);
        GGG(e, a, b, c, d, pBuffer[ 5],  6);
        GGG(d, e, a, b, c, pBuffer[12],  9);
        GGG(c, d, e, a, b, pBuffer[ 2], 12);
        GGG(b, c, d, e, a, pBuffer[13],  9);
        GGG(a, b, c, d, e, pBuffer[ 9], 12);
        GGG(e, a, b, c, d, pBuffer[ 7],  5);
        GGG(d, e, a, b, c, pBuffer[10], 15);
        GGG(c, d, e, a, b, pBuffer[14],  8);
    }

    static inline void HHHR(DWord& a, DWord& b, DWord& c, DWord& d, DWord& e, DWord* pBuffer)
    {
        HHH(d, e, a, b, c, pBuffer[15],  9);
        HHH(c, d, e, a, b, pBuffer[ 5],  7);
        HHH(b, c, d, e, a, pBuffer[ 1], 15);
        HHH(a, b, c, d, e, pBuffer[ 3], 11);
        HHH(e, a, b, c, d, pBuffer[ 7],  8);
        HHH(d, e, a, b, c, pBuffer[14],  6);
        HHH(c, d, e, a, b, pBuffer[ 6],  6);
        HHH(b, c, d, e, a, pBuffer[ 9], 14);
        HHH(a, b, c, d, e, pBuffer[11], 12);
        HHH(e, a, b, c, d, pBuffer[ 8], 13);
        HHH(d, e, a, b, c, pBuffer[12],  5);
        HHH(c, d, e, a, b, pBuffer[ 2], 14);
        HHH(b, c, d, e, a, pBuffer[10], 13);
        HHH(a, b, c, d, e, pBuffer[ 0], 13);
        HHH(e, a, b, c, d, pBuffer[ 4],  7);
        HHH(d, e, a, b, c, pBuffer[13],  5);
    }

    static inline void IIIR(DWord& a, DWord& b, DWord& c, DWord& d, DWord& e, DWord* pBuffer)
    {
        III(e, a, b, c, d, pBuffer[ 6],  9); 
        III(d, e, a, b, c, pBuffer[11], 13);
        III(c, d, e, a, b, pBuffer[ 3], 15);
        III(b, c, d, e, a, pBuffer[ 7],  7);
        III(a, b, c, d, e, pBuffer[ 0], 12);
        III(e, a, b, c, d, pBuffer[13],  8);
        III(d, e, a, b, c, pBuffer[ 5],  9);
        III(c, d, e, a, b, pBuffer[10], 11);
        III(b, c, d, e, a, pBuffer[14],  7);
        III(a, b, c, d, e, pBuffer[15],  7);
        III(e, a, b, c, d, pBuffer[ 8], 12);
        III(d, e, a, b, c, pBuffer[12],  7);
        III(c, d, e, a, b, pBuffer[ 4],  6);
        III(b, c, d, e, a, pBuffer[ 9], 15);
        III(a, b, c, d, e, pBuffer[ 1], 13);
        III(e, a, b, c, d, pBuffer[ 2], 11);
    }

    static inline void JJJR(DWord& a, DWord& b, DWord& c, DWord& d, DWord& e, DWord* pBuffer)
    {
        JJJ(a, b, c, d, e, pBuffer[ 5],  8);
        JJJ(e, a, b, c, d, pBuffer[14],  9);
        JJJ(d, e, a, b, c, pBuffer[ 7],  9);
        JJJ(c, d, e, a, b, pBuffer[ 0], 11);
        JJJ(b, c, d, e, a, pBuffer[ 9], 13);
        JJJ(a, b, c, d, e, pBuffer[ 2], 15);
        JJJ(e, a, b, c, d, pBuffer[11], 15);
        JJJ(d, e, a, b, c, pBuffer[ 4],  5);
        JJJ(c, d, e, a, b, pBuffer[13],  7);
        JJJ(b, c, d, e, a, pBuffer[ 6],  7);
        JJJ(a, b, c, d, e, pBuffer[15],  8);
        JJJ(e, a, b, c, d, pBuffer[ 8], 11);
        JJJ(d, e, a, b, c, pBuffer[ 1], 14);
        JJJ(c, d, e, a, b, pBuffer[10], 14);
        JJJ(b, c, d, e, a, pBuffer[ 3], 12);
        JJJ(a, b, c, d, e, pBuffer[12],  6);
    }

};

class CRMD160Algorithm : public CRMDBase<5>, public CRMD160Base {
public:
    virtual const char* GetAlgName() { return "RIPEMD-160";}
    virtual const char* GetImpAuthor() { return "Programming Unlimited";}
    virtual int  GetImpVersion() { return 1000000;}
    virtual const char* GetAdditionalInfo() { return "A digest algorithm developed in Europe by Hans Dobbertin, "
                                                     "Antoon Bosserlaers, and Bart Preneel.  Has similarities with "
                                                     "MD5 and SHA-1";}


    virtual void Initialize()
    {
        m_nResult[0] = 0x67452301;
        m_nResult[1] = 0xefcdab89;
        m_nResult[2] = 0x98badcfe;
        m_nResult[3] = 0x10325476;
        m_nResult[4] = 0xc3d2e1f0;

        m_nBits[0] = 0;
        m_nBits[1] = 0;
    }

    virtual void Transform()
    {
        register DWord a, b, c, d, e, aa, bb, cc, dd, ee;

        a = aa = m_nResult[0];
        b = bb = m_nResult[1];
        c = cc = m_nResult[2];
        d = dd = m_nResult[3];
        e = ee = m_nResult[4];

        FFR(a, b, c, d, e, m_pBuffer);
        GGR(a, b, c, d, e, m_pBuffer);
        HHR(a, b, c, d, e, m_pBuffer);
        IIR(a, b, c, d, e, m_pBuffer);
        JJR(a, b, c, d, e, m_pBuffer);
        JJJR(aa, bb, cc, dd, ee, m_pBuffer);
        IIIR(aa, bb, cc, dd, ee, m_pBuffer);
        HHHR(aa, bb, cc, dd, ee, m_pBuffer);
        GGGR(aa, bb, cc, dd, ee, m_pBuffer);
        FFFR(aa, bb, cc, dd, ee, m_pBuffer);

        dd += c + m_nResult[1];
        m_nResult[1] = m_nResult[2] + d + ee;
        m_nResult[2] = m_nResult[3] + e + aa;
        m_nResult[3] = m_nResult[4] + a + bb;
        m_nResult[4] = m_nResult[0] + b + cc;
        m_nResult[0] = dd;

        return;
    }

};

