////////////////////////////////////////////////////////////////////////
//                                                                    //
// MicrOTP: A tool for computing one time passwords for S/Key logins. //
//                                                                    //
// (C) 2008 Tom Bramer                                                //
//                                                                    //
// This program is free software; you can redistribute it and/or      //
// modify it under the terms of the GNU General Public License        //
// as published by the Free Software Foundation; either version 2     //
// of the License, or (at your option) any later version.             //
//                                                                    //
// This program is distributed in the hope that it will be useful,    //
// but WITHOUT ANY WARRANTY; without even the implied warranty of     //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      //
// GNU General Public License for more details.                       //
//                                                                    //
// You should have received a copy of the GNU General Public License  //
// along with this program; if not, write to the Free Software        //
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA      //
// 02110-1301, USA.                                                   //
//                                                                    //
////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <fv/config.h>
#endif

#include <string>
#include <vector>
#include <windows.h>
#include <tchar.h>
#include <fv/defs.hpp>
#include <fv/types.hpp>
#include <fv/hashiface.hpp>
#include <fv/build.hpp>
#include <fv/version.hpp>
#include "common.hpp"

USING_NAMESPACE(PU::FV);

typedef unsigned short Word;
typedef unsigned long DWord;

#ifdef _MSC_VER
typedef unsigned __int64 QWord;
#define LL(k) k ## i64
#else
typedef unsigned long long QWord;
#define LL(k) k ## ULL
#endif


#ifdef BIG_ENDIAN  /* Assume LITTLE_ENDIAN */

template <typename T>
class RMDByteReverse {
public:
    static inline void Do(Byte* buffer, DWord len) { return ;}
};

#else

template <typename T>
class SHAByteReverse {
};

template <>
class SHAByteReverse<DWord> {
public:
    static inline void Do(Byte* buffer, DWord len)
    {
        do {
            DWord temp =
            (((DWord) buffer[0] << 8) | buffer[1]) << 16 |
            (((DWord) buffer[2] << 8) | buffer[3]);
            *(DWord*) buffer = temp;
            buffer += 4;
        } while (--len);
        return;
    }
};

template <>
class SHAByteReverse<QWord> {
public:
    static inline void Do(Byte* buffer, DWord len)
    {
        do {
            QWord temp =
            (QWord) buffer[0] << 56 |
            (QWord) buffer[1] << 48 |
            (QWord) buffer[2] << 40 |
            (QWord) buffer[3] << 32 |
            (QWord) buffer[4] << 24 |
            (QWord) buffer[5] << 16 |
            (QWord) buffer[6] << 8  |
            (QWord) buffer[7];
            *(QWord*) buffer = temp;
            buffer += 8;
        } while (--len);
        return;
    }
};

#endif



template <typename BlockType, int HashSize, int BufferSize>
class CSHABase : public CHashAlgorithm {
public:
	virtual const char* GetAlgName() = 0;
	virtual const char* GetImpAuthor() = 0;
	virtual int  GetImpVersion() = 0;
	virtual const char* GetAdditionalInfo() = 0;
	virtual void Transform() = 0;
    virtual void Initialize() = 0;

    CSHABase() { m_pBuffer = reinterpret_cast<BlockType*>(m_aBuffer);}

    static inline BlockType ROTL(BlockType X, BlockType Y) 
    { return(X << Y) | (X >> ((sizeof(BlockType) * 8) - Y));}
    static inline BlockType ROTR(BlockType X, BlockType Y) 
    { return(X >> Y) | (X << ((sizeof(BlockType) * 8) - Y));}

    virtual void DoStep(Byte* pBuffer, BufferLength nLength)
    {
        BlockType temp, temp2 = nLength;

        temp = m_nBits[0];
        if ((m_nBits[0] = temp + ((BlockType) nLength << 3)) < temp)
            m_nBits[1]++;
        m_nBits[1] += temp2 >> ((8 * sizeof(BlockType)) - 3);
        temp = (temp >> 3) & ((BufferSize * sizeof(BlockType)) - 1);

        if (temp) {  /* Handle buffered data */
            Byte* bp = m_aBuffer + temp;
            temp = ((BufferSize * sizeof(BlockType)) - 1) - temp;

            if (nLength < temp) {
                memcpy(bp, pBuffer, nLength);
                return;
            }

            memcpy(bp, pBuffer, temp);
            SHAByteReverse<BlockType>::Do(m_aBuffer, BufferSize);
            Transform();
            pBuffer += temp;
            nLength -= temp;
        }

        while (nLength >= (BufferSize * sizeof(BlockType))) {
            memcpy(m_aBuffer, pBuffer, (BufferSize * sizeof(BlockType)));
            SHAByteReverse<BlockType>::Do(m_aBuffer, BufferSize);
            Transform();
            pBuffer += (BufferSize * sizeof(BlockType));
            nLength -= (BufferSize * sizeof(BlockType));
        }

        memcpy(m_aBuffer, pBuffer, nLength);
    }

    virtual void Finalize()
    {
        BlockType count;
        Byte* bp;

        count = (m_nBits[0] >> 3) & ((sizeof(BlockType) * BufferSize) - 1);

        bp = m_aBuffer + count;
        *bp++ = 0x80;

        count = ((sizeof(BlockType) * BufferSize) - 1) - count;

        if (count < (2 * sizeof(BlockType))) {
            memset(bp, 0, count);
            SHAByteReverse<BlockType>::Do(m_aBuffer, BufferSize);
            Transform();
            memset(m_aBuffer, 0, (BufferSize - 2) * sizeof(BlockType));
        } else {
            memset(bp, 0, count - (2 * sizeof(BlockType)));
        }

        ((BlockType*)m_aBuffer)[14] = m_nBits[1];
        ((BlockType*)m_aBuffer)[15] = m_nBits[0];
        SHAByteReverse<BlockType>::Do(m_aBuffer, BufferSize - 2);

        Transform();
        SHAByteReverse<BlockType>::Do((Byte*)m_nResult, HashSize);
        memset(m_aBuffer, 0, sizeof(BlockType) * BufferSize);
        return;
    }

	virtual int GetStringResult(char* pBuffer, int nBufferSize)
	{
		if(nBufferSize == (2 * sizeof(BlockType) * HashSize) + 1) {
			HashToHexStr(pBuffer, (Byte*)m_nResult, sizeof(BlockType) * HashSize);
		}

		return (2 * sizeof(BlockType) * HashSize) + 1;
	}

    virtual int GetByteResult(Byte* pBuffer, int nBufferSize)
    {
        if (nBufferSize == sizeof(BlockType) * HashSize) {
            memcpy(pBuffer, (Byte*)m_nResult, sizeof(BlockType) * HashSize);
        }
        return sizeof(BlockType) * HashSize;
    }

protected:
    BlockType  m_nResult[HashSize];
    BlockType  m_nTemp[HashSize];
    BlockType  m_nBits[2];
    Byte       m_aBuffer[BufferSize * sizeof(BlockType)];
    BlockType* m_pBuffer;
};


class CSHA1Algorithm : public CSHABase<DWord, 5, 16> {
public:
    static const DWord K[4];

    virtual const char* GetAlgName() { return "SHA-1";}
    virtual const char* GetImpAuthor() { return "Programming Unlimited";}
    virtual int  GetImpVersion() { return 1000000;}
    virtual const char* GetAdditionalInfo() { return "SHA-1 algorithm as developed by NIST.";}

    static inline DWord F(DWord X, DWord Y, DWord Z) { return Z ^ (X & (Y ^ Z));}
    static inline DWord G(DWord X, DWord Y, DWord Z) { return X ^ Y ^ Z;}
    static inline DWord H(DWord X, DWord Y, DWord Z) { return(X & Y) | (Z & (X | Y));}
    static inline DWord I(DWord X, DWord Y, DWord Z) { return X ^ Y ^ Z;}

    static inline void FF(DWord R, DWord& S, DWord T, DWord U, DWord& V, DWord M)
    {
        V += ROTL(R, 5) + F(S, T, U) + K[0] + M;
        S = ROTL(S, 30);
    }

    static inline void GG(DWord R, DWord& S, DWord T, DWord U, DWord& V, DWord M)
    {
        V += ROTL(R, 5) + G(S, T, U) + K[1] + M;
        S = ROTL(S, 30);
    }

    static inline void HH(DWord R, DWord& S, DWord T, DWord U, DWord& V, DWord M)
    {
        V += ROTL(R, 5) + H(S, T, U) + K[2] + M;
        S = ROTL(S, 30);
    }

    static inline void II(DWord R, DWord& S, DWord T, DWord U, DWord& V, DWord M)
    {
        V += ROTL(R, 5) + I(S, T, U) + K[3] + M;
        S = ROTL(S, 30);
    }

    static inline DWord SEL(DWord* X, DWord Y)
    {
        return(X[Y & 0x0F] = ROTL(X[Y & 0x0F] ^ 
                                  X[(Y - 3) & 0x0F] ^ 
                                  X[(Y - 8) & 0x0F] ^ 
                                  X[(Y - 14) & 0x0F], 1));
    }

    virtual void Initialize()
    {
        m_nResult[0] = 0x67452301;
        m_nResult[1] = 0xefcdab89;
        m_nResult[2] = 0x98badcfe;
        m_nResult[3] = 0x10325476;
        m_nResult[4] = 0xc3d2e1f0;

        m_nBits[0] = 0;
        m_nBits[1] = 0;
    }

    virtual void Transform()
    {
        register DWord a, b, c, d, e;

        a = m_nResult[0];
        b = m_nResult[1];
        c = m_nResult[2];
        d = m_nResult[3];
        e = m_nResult[4];

        FF(a, b, c, d, e, m_pBuffer[0 ]);
        FF(e, a, b, c, d, m_pBuffer[1 ]);
        FF(d, e, a, b, c, m_pBuffer[2 ]);
        FF(c, d, e, a, b, m_pBuffer[3 ]);
        FF(b, c, d, e, a, m_pBuffer[4 ]);
        FF(a, b, c, d, e, m_pBuffer[5 ]);
        FF(e, a, b, c, d, m_pBuffer[6 ]);
        FF(d, e, a, b, c, m_pBuffer[7 ]);
        FF(c, d, e, a, b, m_pBuffer[8 ]);
        FF(b, c, d, e, a, m_pBuffer[9 ]);
        FF(a, b, c, d, e, m_pBuffer[10]);
        FF(e, a, b, c, d, m_pBuffer[11]);
        FF(d, e, a, b, c, m_pBuffer[12]);
        FF(c, d, e, a, b, m_pBuffer[13]);
        FF(b, c, d, e, a, m_pBuffer[14]);
        FF(a, b, c, d, e, m_pBuffer[15]);
        FF(e, a, b, c, d, SEL(m_pBuffer, 16));
        FF(d, e, a, b, c, SEL(m_pBuffer, 17));
        FF(c, d, e, a, b, SEL(m_pBuffer, 18));
        FF(b, c, d, e, a, SEL(m_pBuffer, 19));

        GG(a, b, c, d, e, SEL(m_pBuffer, 20));
        GG(e, a, b, c, d, SEL(m_pBuffer, 21));
        GG(d, e, a, b, c, SEL(m_pBuffer, 22));
        GG(c, d, e, a, b, SEL(m_pBuffer, 23));
        GG(b, c, d, e, a, SEL(m_pBuffer, 24));
        GG(a, b, c, d, e, SEL(m_pBuffer, 25));
        GG(e, a, b, c, d, SEL(m_pBuffer, 26));
        GG(d, e, a, b, c, SEL(m_pBuffer, 27));
        GG(c, d, e, a, b, SEL(m_pBuffer, 28));
        GG(b, c, d, e, a, SEL(m_pBuffer, 29));
        GG(a, b, c, d, e, SEL(m_pBuffer, 30));
        GG(e, a, b, c, d, SEL(m_pBuffer, 31));
        GG(d, e, a, b, c, SEL(m_pBuffer, 32));
        GG(c, d, e, a, b, SEL(m_pBuffer, 33));
        GG(b, c, d, e, a, SEL(m_pBuffer, 34));
        GG(a, b, c, d, e, SEL(m_pBuffer, 35));
        GG(e, a, b, c, d, SEL(m_pBuffer, 36));
        GG(d, e, a, b, c, SEL(m_pBuffer, 37));
        GG(c, d, e, a, b, SEL(m_pBuffer, 38));
        GG(b, c, d, e, a, SEL(m_pBuffer, 39));

        HH(a, b, c, d, e, SEL(m_pBuffer, 40));
        HH(e, a, b, c, d, SEL(m_pBuffer, 41));
        HH(d, e, a, b, c, SEL(m_pBuffer, 42));
        HH(c, d, e, a, b, SEL(m_pBuffer, 43));
        HH(b, c, d, e, a, SEL(m_pBuffer, 44));
        HH(a, b, c, d, e, SEL(m_pBuffer, 45));
        HH(e, a, b, c, d, SEL(m_pBuffer, 46));
        HH(d, e, a, b, c, SEL(m_pBuffer, 47));
        HH(c, d, e, a, b, SEL(m_pBuffer, 48));
        HH(b, c, d, e, a, SEL(m_pBuffer, 49));
        HH(a, b, c, d, e, SEL(m_pBuffer, 50));
        HH(e, a, b, c, d, SEL(m_pBuffer, 51));
        HH(d, e, a, b, c, SEL(m_pBuffer, 52));
        HH(c, d, e, a, b, SEL(m_pBuffer, 53));
        HH(b, c, d, e, a, SEL(m_pBuffer, 54));
        HH(a, b, c, d, e, SEL(m_pBuffer, 55));
        HH(e, a, b, c, d, SEL(m_pBuffer, 56));
        HH(d, e, a, b, c, SEL(m_pBuffer, 57));
        HH(c, d, e, a, b, SEL(m_pBuffer, 58));
        HH(b, c, d, e, a, SEL(m_pBuffer, 59));

        II(a, b, c, d, e, SEL(m_pBuffer, 60));
        II(e, a, b, c, d, SEL(m_pBuffer, 61));
        II(d, e, a, b, c, SEL(m_pBuffer, 62));
        II(c, d, e, a, b, SEL(m_pBuffer, 63));
        II(b, c, d, e, a, SEL(m_pBuffer, 64));
        II(a, b, c, d, e, SEL(m_pBuffer, 65));
        II(e, a, b, c, d, SEL(m_pBuffer, 66));
        II(d, e, a, b, c, SEL(m_pBuffer, 67));
        II(c, d, e, a, b, SEL(m_pBuffer, 68));
        II(b, c, d, e, a, SEL(m_pBuffer, 69));
        II(a, b, c, d, e, SEL(m_pBuffer, 70));
        II(e, a, b, c, d, SEL(m_pBuffer, 71));
        II(d, e, a, b, c, SEL(m_pBuffer, 72));
        II(c, d, e, a, b, SEL(m_pBuffer, 73));
        II(b, c, d, e, a, SEL(m_pBuffer, 74));
        II(a, b, c, d, e, SEL(m_pBuffer, 75));
        II(e, a, b, c, d, SEL(m_pBuffer, 76));
        II(d, e, a, b, c, SEL(m_pBuffer, 77));
        II(c, d, e, a, b, SEL(m_pBuffer, 78));
        II(b, c, d, e, a, SEL(m_pBuffer, 79));

        m_nResult[0] += a;
        m_nResult[1] += b;
        m_nResult[2] += c;
        m_nResult[3] += d;
        m_nResult[4] += e;

        return;
    }

};

const DWord CSHA1Algorithm::K[4] = { 0x5A827999, 0x6ED9EBA1, 0x8F1BBCDC, 0xCA62C1D6};

